(*
 * This file is part of Bolt.
 * Copyright (C) 2009-2012 Xavier Clerc.
 *
 * Bolt is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * Bolt is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *)


let register_logger normalized_name level filter layout output (output_name, output_rotate) =
  let layout = lazy (try Layout.get layout with Not_found -> Layout.default) in
  let filter = lazy (try Filter.get filter with Not_found -> Filter.all) in
  let output = lazy
      (try
        (Output.get output) output_name output_rotate layout
      with Not_found -> Output.void output_name output_rotate layout) in
  let infos = { Tree.name = normalized_name;
                level; filter; layout; output; } in
  Tree.register_logger infos

let register name level filter layout output output_params =
  let normalized_name = Name.of_string name in
  register_logger normalized_name level filter layout output output_params

let log_event name level file line column properties error msg =
  let orig_event =
    Event.make
      name
      level
      ~file:file
      ~line:line
      ~column:column
      ~properties:properties
      ~error:error
      msg in
  let loggers = Tree.get_loggers name in
  List.iter
    (fun (nam, lst) ->
      let event = Event.with_logger nam orig_event in
      List.iter
        (fun logger ->
          try
            let _, _, layout = Lazy.force logger.Tree.layout in
            let output = Lazy.force logger.Tree.output in
            let filter = Lazy.force logger.Tree.filter in
            if (level <= logger.Tree.level) && (filter event) then
              output#write (layout event)
          with _ -> ())
        lst)
    loggers

let log name level ?(file="") ?(line=(-1)) ?(column=(-1)) ?(properties=[]) ?(error=None) msg =
  let name = Name.of_string name in
  log_event name level file line column properties error msg

let prepare name =
  let name = Name.of_string name in
  Tree.make_node name

let configuration =
  let get_from_env =
    try
      (Sys.getenv "BOLT_CONFIG"), ConfigurationNew.load
    with _ ->
      (Sys.getenv "BOLT_FILE"), ConfigurationOld.load in
  try
    let file, func = get_from_env in
    Some (func file)
  with
  | Not_found -> None
  | Configuration.Exception (line, err) ->
      let msg =
        Printf.sprintf "syntax error in configuration file at line %d:\n  %s"
          line
          err in
      Utils.verbose msg;
      None
  | _ ->
      Utils.verbose "unable to load configuration";
      None

let () =
  try
    let plugins_env = Sys.getenv "BOLT_PLUGINS" in
    let plugins_list = Utils.split "," plugins_env in
    let plugins_list = List.map Utils.trim plugins_list in
    List.iter Dynlink.loadfile_private plugins_list
  with _ -> ()

let signal_of_string s =
  match String.lowercase s with
  | "sighup" -> Sys.sighup
  | "sigusr1" -> Sys.sigusr1
  | "sigusr2" -> Sys.sigusr2
  | _ -> failwith "unsupported signal"

let () =
  let read_lines file =
    try
      let ch = open_in file in
      let res = ref [] in
      (try
        while true do
          res := (input_line ch) :: !res
        done
      with End_of_file -> ());
      List.rev !res
    with _ -> [] in
  match configuration with
  | Some conf ->
    List.iter
      (fun section ->
        let assoc x = List.assoc x section.Configuration.elements in
        let level = try Level.of_string (assoc "level") with _ -> Level.FATAL in
        let filter = try assoc "filter" with _ -> "all" in
        let layout =
          try
            match assoc "layout" with
            | "pattern" ->
                let header = try read_lines (assoc "pattern-header-file") with _ -> [] in
                let footer = try read_lines (assoc "pattern-footer-file") with _ -> [] in
                Layout.register_unnamed (Layout.pattern header footer (assoc "pattern"))
            | "csv" ->
                let sep = try assoc "csv-separator" with _ -> ";" in
                let elems = Utils.split " \t" (assoc "csv-elements") in
                Layout.register_unnamed (Layout.csv sep elems)
            | x -> x
          with _ -> "default" in
        let output = try assoc "output" with _ -> "file" in
        let name = try assoc "name" with _ -> "<stderr>" in
        let seconds = try Some (float_of_string (assoc "rotate")) with _ -> None in
        let signal = try Some (signal_of_string (assoc "signal")) with _ -> None in
        (match signal with
        | Some s ->
            let f _ =
              Output.signal := true in
            Sys.set_signal s (Sys.Signal_handle f)
        | None -> ());
        let rotate = { Output.seconds_elapsed = seconds;
                       Output.signal_caught = signal; } in
        register_logger
          section.Configuration.name
          level
          filter
          layout
          output
          (name, rotate))
      conf
  | None ->
      ()
