(*
 * This file is part of Bolt.
 * Copyright (C) 2009-2012 Xavier Clerc.
 *
 * Bolt is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * Bolt is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *)

(** This module provides support for the Pajé tool.

    Pajé (available at {i https://gforge.inria.fr/projects/paje/}) is
    a graphical tool for the analysis of multithreads and/or multiprocesses
    programs.

    Bolt can produce Pajé-comptabile traces by using statements like:
    {C [LOG Paje.t WITH Paje.new_event ...;] }
    more information can be found in the Bolt manual. *)


(** {6 Definitions} *)

val t : string
(** The identifier message for Paje event. *)

type properties = (string * string) list
(** The type of information to be recorded in a trace. *)


(** {6 Predefined events: trace definition} *)

val define_container_type : name:string -> ?alias:string -> ?container_type:string -> properties -> properties
(** Defines a new type of container with given name, alias, parent
    type, and a list of additional properties. *)

val define_event_type : name:string -> ?alias:string -> container_type:string -> properties -> properties
(** Defines a new type of event with name, alias, container type, and a
    list of additional properties. Events are displayed on time line
    through a mark such as a triangle. *)

val define_state_type : name:string -> ?alias:string -> container_type:string -> properties -> properties
(** Defines a new type of state with name, alias, container type, and a
    list of additonnal properties. States are displayed on time line
    through colored rectangles. *)

val define_variable_type : name:string -> ?alias:string -> container_type:string -> properties -> properties
(** Defines a new type of variable with name, alias, container type, and a
    list of additonnal properties. States are displayed on time line
    through float values. *)

val define_link_type : name:string -> ?alias:string -> container_type:string -> source_container_type:string -> dest_container_type:string -> properties -> properties
(** Defines a new type of link with name, alias, container type, both
    source and destination container types, and a list of additonnal
    properties. States are displayed on time line through arrows. *)

val define_entity_value : name:string -> ?alias:string -> entity_type:string -> properties -> properties
(** Defines a new value for an entity with name, alias, entity type,
    and a list of additional properties. *)


(** {6 Predefined events: trace recording} *)

val create_container : name:string -> ?alias:string -> typ:string -> ?container:string -> properties -> properties
(** Creates a container with name, alias, type, parent and a list of
    additional properties. *)

val destroy_container : name:string -> typ:string -> properties -> properties
(** Destroys a container with name, type, and a list of additional
    properties. *)

val new_event : typ:string -> container:string -> value:string -> properties -> properties
(** Record a new event with type, container where event happens, value
    of event, and a additional list of properties. *)

val set_state : typ:string -> container:string -> value:string -> properties -> properties
(** Changes the state of a given container to a new value with passed
    type, and additional list of properties. *)

val push_state : typ:string -> container:string -> value:string -> properties -> properties
(** Pushes the current state of a given container to its own stack,
    and changes its states to the pass type and value. *)

val pop_state : typ:string -> container:string -> properties -> properties
(** Changes the states of a given container by poping a previously pushed
    value from its own stack. *)

val set_variable : typ:string -> container:string -> value:float -> properties -> properties
(** Sets the value of a container variable. *)

val add_variable : typ:string -> container:string -> value:float -> properties -> properties
(** Increments the value of a container variable by a given amount. *)

val sub_variable : typ:string -> container:string -> value:float -> properties -> properties
(** Decrements the value of a container variable by a given amount. *)

val start_link : typ:string -> container:string -> source_container:string -> value:string -> key:string -> properties -> properties
(** Records the start of a link occuring inside a container, from a source.
    The key/value pair is used to matched with an associated end of link. *)

val end_link : typ:string -> container:string -> dest_container:string -> value:string -> key:string -> properties -> properties
(** Records the end of a link occuring inside a container, to a destination.
    The key/value pair is used to matched with an associated start of link. *)


(** {6 Layout elements} *)

val header : string list
(** The header defining Paje events. *)

val render : Event.t -> string
(** The rendering function for Paje format. *)

val layout : Layout.t
(** The layout supporting the "Pajé" trace format.
    The message of the event should match the name of an {i EventDef}
    defined by the "Pajé trace file format" available at 
    http://sourceforge.net/projects/paje/. The values of the fields
    are taken from the event properties, the ["Time"] field being
    automatically set to the time elapsed since program start (unless
    explicitly set through a property). *)

val layout_noheader : Layout.t
(* Same as [paje], but without any header.
   Useful to avoid duplicate information when several process produce
   Paje traces. *)


(** {6 Functorial interface} *)

module type Definitions = sig

  val logger : string
  (** The name of the logger used to register definitions. *)

  val level : Level.t
  (** The level used to register definitions. *)

  type container_type
  (** The type of containers, typically a sum type. *)

  val container_types : (container_type * string * string * (container_type option)) list
  (** The definition of container types: value, name, alias and optional parent. *)

  type event_type
  (** The type of events, typically a sum type. *)

  val event_types : (event_type * string * string * container_type) list
  (** The definition of event types: value, name, alias and container. *)

  type state_type
  (** The type of states, typically a sum type. *)

  val state_types : (state_type * string * string * container_type) list
  (** The definition of event types: value, name, alias and container. *)

  type variable_type
  (** The type of states, typically a sum type. *)

  val variable_types : (variable_type * string * string * container_type) list
  (** The definition of variable types: value, name, alias and container. *)

  type link_type
  (** The type of links, typically a sum type. *)

  val link_types : (link_type * string * string * container_type * container_type * container_type) list
  (** The definition of link types: value, name, alias, parent container, source container, and destination container. *)
  
end
(** The input signature of {!Paje.Make}, that is static definition of
    Pajé trace. *)

module type S = sig
  val t : string
  type properties = (string * string) list
  type container_type
  type event_type
  type state_type
  type variable_type
  type link_type
  val create_container : name:string -> ?alias:string -> typ:container_type -> ?container:string -> properties -> properties
  val destroy_container : name:string -> typ:container_type -> properties -> properties
  val new_event : typ:event_type -> container:string -> value:string -> properties -> properties
  val set_state : typ:state_type -> container:string -> value:string -> properties -> properties
  val push_state : typ:state_type -> container:string -> value:string -> properties -> properties
  val pop_state : typ:state_type -> container:string -> properties -> properties
  val set_variable : typ:variable_type -> container:string -> value:float -> properties -> properties
  val add_variable : typ:variable_type -> container:string -> value:float -> properties -> properties
  val sub_variable : typ:variable_type -> container:string -> value:float -> properties -> properties
  val start_link : typ:link_type -> container:string -> source_container:string -> value:string -> key:string -> properties -> properties
  val end_link : typ:link_type -> container:string -> dest_container:string -> value:string -> key:string -> properties -> properties
end
(** The output signature of {!Paje.Make}, that is type-safe interface
    to Pajé functions. *)

module Make (D : Definitions) : S
  with type container_type = D.container_type
  and type event_type = D.event_type
  and type state_type = D.state_type
  and type variable_type = D.variable_type
  and type link_type = D.link_type
(** Functor building a type-safe version of Pajé trace-recording functions.
    Definitions from the passed module are recorded at the "TRACE" level. *)
